/*
 * Copyright 2018, Theodore Chabardes
 * <theodore.chabardes@mines-paristech.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef _CONCURRENT_LINKED_LIST_H_
#define _CONCURRENT_LINKED_LIST_H_

#include "hierarchical_queue.hpp"
#ifdef _OPENMP
#include <omp.h>
#endif // _OPENMP
#include <vector>

template <class T> class concurrent_linked_list {
private:
  struct node {
    T data;
    node *next;
    node *prev;
  };

public:
  concurrent_linked_list();
  ~concurrent_linked_list();
  concurrent_linked_list(const concurrent_linked_list<T> &) = delete;

  // Synchronised operations.
  void push_back_sync(T d);
  void fuse_sync(concurrent_linked_list<T> *ll);
  template <class T2, class Compare>
  void remove_and_queue_each_element_sync(Compare comp, std::vector<T> &q,
                                          T2 *heights, T2 param);

  // No Synchronisation.
  void remove_all();
  size_t size() { return count; }
  void print_self();
  template <class T2> T2 get_lowest_value_nsync(T2 *heights);
  template <class T2, class Compare>
  void remove_and_queue_each_element_nsync(Compare comp, std::queue<T> &q,
                                           T2 *heights, T2 param);
  template <class T2, class Compare>
  void remove_and_queue_each_element_nsync(Compare comp,
                                           hierarchical_queue<T2> &q,
                                           T2 *heights, T2 param);
  void remove_queue_all_elements_nsync(std::queue<T> &);
  void remove_queue_all_elements_nsync(std::vector<T> &);
  template <class T2>
  void remove_queue_all_elements_nsync(hierarchical_queue<T2> &, T2* heights);


private:
  node *newNode(T data) {
    node *result = new node;
    result->data = data;
    result->next = NULL;
    result->prev = NULL;
    return result;
  }
  void detach_data() {
    head = NULL;
    tail = NULL;
    count = 0;
  }

  node *&getTail() { return tail; }
  node *&getHead() { return head; }
#ifdef _OPENMP
  omp_lock_t &getLock() { return writelock; }
#endif // _OPENMP

  size_t count;
  node *head;
  node *tail;
#ifdef _OPENMP
  omp_lock_t writelock;
#endif //_OPENMP
};

// Should always be run in isolation
template <class T> concurrent_linked_list<T>::concurrent_linked_list() {
  count = 0;
  head = NULL;
  tail = NULL;
#ifdef _OPENMP
  omp_init_lock(&writelock);
#endif // _OPENMP
}

// Should always be run in isolation
template <class T> concurrent_linked_list<T>::~concurrent_linked_list() {
  remove_all();
#ifdef _OPENMP
  omp_destroy_lock(&writelock);
#endif // _OPENMP
}

template <class T> void concurrent_linked_list<T>::remove_all() {
  node *n = head, *next;
  size_t i = 0;
  while (i < count) {
    next = n->next;
    delete n;
    n = next;
    ++i;
  }
  detach_data();
}

template <class T> void concurrent_linked_list<T>::print_self() {
  node *n = head;
  size_t i = 0;
  std::cout << (int)count << ": ";
  while (i < count) {
    std::cout << (int)n->data;
    n = n->next;
    ++i;
    if (i < count)
      std::cout << ", ";
  }
  if (head != NULL && tail != NULL)
    std::cout << " (HEAD: " << (int)head->data << ", TAIL: " << (int)tail->data
              << ")";
}

template <class T> void concurrent_linked_list<T>::push_back_sync(T d) {
  node *tmp = newNode(d);

#ifdef _OPENMP
  omp_set_lock(&writelock);
#endif //_OPENMP
  tmp->prev = tail;
  ++count;

  if (count == 1) {
    head = tmp;
    tail = tmp;
  } else {
    tail->next = tmp;
    tail = tmp;
  }
#ifdef _OPENMP
  omp_unset_lock(&writelock);
#endif // _OPENMP
}

template <class T>
template <class T2, class Compare>
void concurrent_linked_list<T>::remove_and_queue_each_element_sync(
    Compare comp, std::vector<T> &q, T2 *heights, T2 param) {
#ifdef _OPENMP
  omp_set_lock(&writelock);
#endif // _OPENMP

  node *n = head, *tmp;
  size_t i = 0;

  while (i < count) {
    tmp = n->next;
    if (comp(heights[n->data], param)) {
      if (n->prev == NULL)
        head = n->next;
      else
        n->prev->next = n->next;

      if (n->next == NULL)
        tail = n->prev;
      else
        n->next->prev = n->prev;

      --count;
      q.push_back(n->data);
      delete n;
    } else
      ++i;
    n = tmp;
  }
#ifdef _OPENMP
  omp_unset_lock(&writelock);
#endif // _OPENMP
}

template <class T>
template <class T2, class Compare>
void concurrent_linked_list<T>::remove_and_queue_each_element_nsync(
    Compare comp, std::queue<T> &q, T2 *heights, T2 param) {

  node *n = head, *tmp;
  size_t i = 0;

  while (i < count) {
    tmp = n->next;
    if (comp(heights[n->data], param)) {
      if (n->prev == NULL)
        head = n->next;
      else
        n->prev->next = n->next;

      if (n->next == NULL)
        tail = n->prev;
      else
        n->next->prev = n->prev;

      --count;
      q.push(n->data);
      delete n;
    } else
      ++i;
    n = tmp;
  }
}

template <class T>
template <class T2, class Compare>
void concurrent_linked_list<T>::remove_and_queue_each_element_nsync(
    Compare comp, hierarchical_queue<T2> &q, T2 *heights, T2 param) {

  node *n = head, *tmp;
  size_t i = 0;

  while (n != NULL) {
    tmp = n->next;
    if (comp(heights[n->data], param)) {
      if (n->prev == NULL)
        head = n->next;
      else
        n->prev->next = n->next;

      if (n->next == NULL)
        tail = n->prev;
      else
        n->next->prev = n->prev;

      --count;
      q.push(heights[n->data], n->data);
      delete n;
    } else
      ++i;
    n = tmp;
  }
}

template <class T>
template <class T2>
T2 concurrent_linked_list<T>::get_lowest_value_nsync(T2 *heights) {
  T2 min_value = std::numeric_limits<T2>::max();
  size_t p;

  node *n = head;
  size_t i = 0;

  while (i < count) {
    p = n->data;
    min_value = std::min(heights[p], min_value);
    n = n->next;
    ++i;
  }
  return min_value;
}

template <class T>
void concurrent_linked_list<T>::remove_queue_all_elements_nsync(
    std::vector<T> &q) {
  node *n = head, *tmp;
  size_t i = 0;

  while (i < count) {
    tmp = n->next;
    q.push_back(n->data);
    delete n;
    n = tmp;
    ++i;
  }

  count = 0;
  head = NULL;
  tail = NULL;
}

template <class T>
void concurrent_linked_list<T>::remove_queue_all_elements_nsync(
    std::queue<T> &q) {
  node *n = head, *tmp;
  size_t i = 0;

  while (i < count) {
    tmp = n->next;
    q.push(n->data);
    delete n;
    n = tmp;
    ++i;
  }

  count = 0;
  head = NULL;
  tail = NULL;
}

template <class T>
template <class T2>
void concurrent_linked_list<T>::remove_queue_all_elements_nsync(
    hierarchical_queue<T2> &q, T2* heights) {
  node *n = head, *tmp;
  size_t i = 0;

  while (i < count) {
    tmp = n->next;
    q.push(heights[n->data], n->data);
    delete n;
    n = tmp;
    ++i;
  }

  count = 0;
  head = NULL;
  tail = NULL;
}

template <class T>
void concurrent_linked_list<T>::fuse_sync(concurrent_linked_list<T> *ll) {
#ifdef _OPENMP
  omp_set_lock(&getLock());
  omp_set_lock(&ll->getLock());
#endif // _OPENMP

  node *headB = ll->getHead();
  node *tailB = ll->getTail();

  if (count == 0) {
    head = headB;
    tail = tailB;
  } else if (ll->size() > 0) {
    tail->next = headB;
    headB->prev = tail;
    tail = tailB;
  }

  count += ll->size();
  ll->detach_data();

#ifdef _OPENMP
  omp_unset_lock(&getLock());
  omp_unset_lock(&ll->getLock());
#endif // _OPENMP
}

#endif // _CONCURRENT_LINKED_LIST_H_
