/*
 * Copyright 2018, Theodore Chabardes
 * <theodore.chabardes@mines-paristech.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef _UNIONFIND_H_
#define _UNIONFIND_H_

#include <atomic>
#include <cstring>
#include <iostream>
#include <limits>
#ifdef _OPENMP
#include <omp.h>
#endif // _OPENMP

// Union-Find by rank algorithm with Path compression, concurrent lock-free
// version.
template <class labelT> class concurrent_union_find {
public:
  concurrent_union_find() {
    set = new std::atomic<uint64_t>[std::numeric_limits<labelT>::max() + 1];
#ifdef _OPENMP
#pragma omp parallel for simd
#endif //_OPENMP
    for (size_t X = 0; X < std::numeric_limits<labelT>::max() + 1; ++X) {
      set[X] = X;
    }
  }

  ~concurrent_union_find() { delete[] set; }

  labelT get_number_of_subset(size_t length) {
    labelT nbr_subset = 0;
    for (uint32_t X = 1; X < (uint32_t)length; ++X) {
      nbr_subset += (Find(X) == X) ? 1 : 0;
    }
    return nbr_subset;
  }

  labelT Find(labelT X) {
    while (X != parent(X)) {
      uint64_t d = set[X];
      uint32_t new_parent = parent((uint32_t)d);
      uint64_t new_d = (d & 0xFFFFFFFF00000000ULL) | new_parent;
      if (d != new_d)
        set[X].compare_exchange_weak(d, new_d);
      X = new_parent;
    }
    return X;
  }

  bool Same(labelT A, labelT B) {

    for (;;) {
      A = Find(A);
      B = Find(B);
      if (A == B)
        return true;
      if (parent(A) == A)
        return false;
    }
  }

  labelT Union(labelT A, labelT B) {

    for (;;) {
      A = Find(A);
      B = Find(B);

      if (A == B)
        return A;

      uint32_t rA = rank(A);
      uint32_t rB = rank(B);

      if (rA > rB || (rA == rB && A < B)) {
        std::swap(rA, rB);
        std::swap(A, B);
      }

      uint64_t d = ((uint64_t)rA << 32) | A;
      uint64_t new_d = ((uint64_t)rA << 32) | B;

      if (!set[(size_t)A].compare_exchange_strong(d, new_d))
        continue;

      if (rA == rB) {
        d = ((uint64_t)rB << 32) | B;
        new_d = ((uint64_t)(rB + 1) << 32) | B;
        set[(size_t)B].compare_exchange_weak(d, new_d);
      }

      break;
    }
    return B;
  }

private:
  uint32_t rank(uint32_t X) const {
    return (uint32_t)((set[(size_t)X].load() >> 32)) & 0x7FFFFFFFu;
  }
  uint32_t parent(uint32_t X) const { return (uint32_t)set[(size_t)X].load(); }

  std::atomic<uint64_t> *set;
};

#endif // _UNIONFIND_H_
