/*
 * Copyright 2018, Theodore Chabardes
 * <theodore.chabardes@mines-paristech.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef _DISTANCE_H_
#define _DISTANCE_H_

#include "connexity.hpp"
#include "neighbors.hpp"
#include <queue>

// Compute the geodesic distance on plateaus.
// Warning: this function does not check for overflow.
template <class T>
void geodesic_distance(const T *in, std::queue<size_t> &q1, unsigned short *out, const domain &d,
                 const structuring_element &se) {
  typedef unsigned short ushort;
  std::queue<size_t> q2;
  std::swap(q1,q2);

  size_t p;
  ushort atomic_lock;
  ushort min_dist;

  do {
    // Swapping while allowing only one instance of pixel per thread...
    while (!q2.empty()) {
      p = q2.front();
      q2.pop();
      ITERATE_NEIGHBORS_BEGIN(p, d, se) {
        if (in[p] == in[_p] && out[_p] > out[p] + 1)
          q1.push(_p);
      }
      ITERATE_NEIGHBORS_END
    }

    while (!q1.empty()) {
      p = q1.front();
      q1.pop();
      min_dist = std::numeric_limits<ushort>::max();

      // We compute the smallest neighbor.
      ITERATE_NEIGHBORS_BEGIN(p, d, se) {
        if (in[p] == in[_p] && out[_p] < out[p])
          min_dist = std::min(out[_p], min_dist);
      }
      ITERATE_NEIGHBORS_END

// Only one instance of p for each thread.
#ifdef _OPENMP
#pragma omp atomic capture
#endif // _OPENMP
      {
        atomic_lock = out[p];
        out[p] = min_dist + 1;
      }
      if (atomic_lock != min_dist + 1) {
        q2.push(p);
      }
    }
  } while (!q2.empty());
}

#endif // _DISTANCE_H_
