/*
 * Copyright 2018, Theodore Chabardes
 * <theodore.chabardes@mines-paristech.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef _FLOOD_UP_H_
#define _FLOOD_UP_H_

#include "buttonhole.hpp"
#include "connexity.hpp"
#include "hierarchical_queue.hpp"
#include "neighbors.hpp"
#include "rules.hpp"
#include <cstdlib>
#include <queue>
typedef unsigned short ushort;

template <class T, class labelT>
void flood_point(size_t p, BHData<ushort> &buttonholes, const T *in,
                unsigned short *homotopy, uint8_t *counts, ushort *bh_sets,
                labelT *out, const domain &d, const structuring_element &se,
                std::vector<size_t> &to_be_queued,
                bool (*func_ptr)(size_t, size_t, const T *, const unsigned short *)) {

  T atomic_count;

  ITERATE_NEIGHBORS_BEGIN(p, d, se) {
    atomic_count = 1;

    if (func_ptr(p, _p, in, homotopy)) {
#ifdef _OPENMP
#pragma omp atomic capture
#endif // _OPENMP
      atomic_count = --counts[_p];
    }

    if (atomic_count == 0) {
      process_point(_p, buttonholes, in, homotopy, bh_sets, out, d, se);
      to_be_queued.push_back(_p);
    }
    
  }
  ITERATE_NEIGHBORS_END
}

template <class T, class labelT>
void process_point(size_t p, BHData<ushort> &buttonholes, const T *in,
                  unsigned short *homotopy, ushort *bh_sets, labelT *out,
                  const domain &d, const structuring_element &se) {
  uint8_t CQl;
  labelT Ql;

  get_Ql (p, in, homotopy, out, d, se, CQl, Ql, lower_neighbor);

  if (!has_beta_pred (p, in, homotopy, out, d, se)) {

    // This applies the rules (12a), (12b) and (12c)
    if (CQl == 1) {
      out [p] = Ql;  
    } else if (CQl > 1) {
      out [p] = WS;
    } else { // CQl == 0
      out [p] = BETA;
    }
  } else {
      out [p] = BETA;
  }

  if (out[p] == BETA) {
    set_buttonhole (p, buttonholes, in, homotopy, bh_sets, d, se);

    // Buttonhole can be flooded from a lower_neighbor or a level neighbor (see paper, section 'button holes').
    // If there is one and only one lower neighbor with a label, then p is a BCL.
    if (CQl == 1) {
      buttonholes.BCL[bh_sets[p]]->push_back_sync(p);
    } else if (CQl == 0) {
      // p is also a BCL if there is only one level neighbor with a label.
      get_Ql (p, in, homotopy, out, d, se, CQl, Ql, lower_neighbor_abs);
      if (CQl == 1) {
        buttonholes.BCL[bh_sets[p]]->push_back_sync(p);
      }
    }
  }
}


// Each buttonhole is flooded separately, because they are divided into
// none-intersecting subsets of points. The flooding is performed as in the
// hierarchical flooding method.
template <class T, class labelT>
void flood_buttonhole(ushort root, BHData<ushort> &buttonholes, T *in,
                     unsigned short *homotopy, ushort *bh_sets, labelT *out, const domain &d,
                     const structuring_element &se) {
  hierarchical_queue<T> BCL;

  buttonholes.BCL[root]->remove_queue_all_elements_nsync(BCL, in);

  while (!BCL.empty()) {
    T level = BCL.get_next_level ();
    process_BH_level (level, BCL, root, buttonholes, in, homotopy, bh_sets, out, d, se); 
  }
}

// Buttonholes are flooded level by level. The BCL points can be of the same
// height but different geodesic distance within a plateau. 
template <class T, class labelT>
void process_BH_level(T level, hierarchical_queue<T> &BCL, ushort root,
                       BHData<ushort> &buttonholes, T *in, unsigned short *homotopy,
                       ushort *bh_sets, labelT *out, const domain &d,
                       const structuring_element &se) {

  uint8_t CQl;
  labelT Ql;
  hierarchical_queue<unsigned short> hq;
  size_t p;

  // Queueing the BCL points...
  while (!BCL.empty() && BCL.get_next_level() == level) {
    p = BCL.pop();
    bh_sets[p] = 0;
    hq.push(homotopy[p], p);
  }

  // Main propagation loop
  while (!hq.empty()) {
    p = hq.pop();

    // Assigning the final label to the p.
    get_Ql (p, in, homotopy, out, d, se, CQl, Ql, lower_neighbor);
    if (CQl == 1) 
      out[p] = Ql;
    else {
      get_Ql (p, in, homotopy, out, d, se, CQl, Ql, lower_neighbor_abs);
      if (CQl == 1)
        // This part never applies to points other than the BCLs due to the modification of height at lines 160-161... 
        // Maybe the label of the BCLs should be computed beforehand and remove this part.
        out[p] = Ql;
      else
        out[p] = WS;
    }

    // If it is not a WS then it propagates to its neighbors.
    if (out[p] != WS) {
      ITERATE_NEIGHBORS_BEGIN(p, d, se) {
        if (buttonholes.set.Same(root, bh_sets[_p])) {
          bh_sets[_p] = 0;
          // If it is higher, it will be held in a next batch.
          if (in[_p] > in[p]) {
            BCL.push(in[_p], _p);
          } else {
            in[_p] = in[p];
            homotopy[_p] = homotopy[p] + 1;
            // Queueing the neighbor.
            hq.push(homotopy[_p], _p);
          }
        }
      }
      ITERATE_NEIGHBORS_END
    }
  }
}

#endif // _FLOOD_UP_H_
