/*
 * copyright 2018, theodore chabardes
 * <theodore.chabardes@mines-paristech.fr>
 *
 * this program is free software: you can redistribute it and/or modify
 * it under the terms of the gnu general public license as published by
 * the free software foundation, either version 3 of the license, or
 * (at your option) any later version.
 * this program is distributed in the hope that it will be useful,
 * but without any warranty; without even the implied warranty of
 * merchantability or fitness for a particular purpose.  see the
 * gnu general public license for more details.
 * you should have received a copy of the gnu general public license
 * along with this program.  if not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef _HIERARCHICAL_QUEUE_H_
#define _HIERARCHICAL_QUEUE_H_

#include <cstddef>
#include <iostream>
#include <limits>
#include <queue>
#ifdef BENCH
#include "bench.hpp"
#endif // BENCH

template <class T1> class hierarchical_queue {
public:
  hierarchical_queue();
  ~hierarchical_queue();
  void push(T1 level, size_t offset);
  size_t pop();
  bool empty();
  T1 get_next_level();

private:
  std::queue<size_t> *queues;
  T1 current_level;
};

template <class T1> hierarchical_queue<T1>::hierarchical_queue() {
  queues = new std::queue<size_t>[std::numeric_limits<T1>::max() + 1];
  current_level = std::numeric_limits<T1>::max();
}

template <class T1> hierarchical_queue<T1>::~hierarchical_queue() {
  delete[] queues;
}

template <class T1> void hierarchical_queue<T1>::push(T1 level, size_t offset) {
  if (level < current_level) {
    current_level = level;
  }
  queues[level].push(offset);
}

template <class T1> size_t hierarchical_queue<T1>::pop() {
  size_t r = 0;
  if (!empty()) {
    r = queues[current_level].front();
    queues[current_level].pop();
  }
  return r;
}

template <class T1> T1 hierarchical_queue<T1>::get_next_level() {
  T1 index = current_level;
  while (index < std::numeric_limits<T1>::max() && queues[index].empty())
    ++index;
  current_level = index;
  return index;
}

template <class T1> bool hierarchical_queue<T1>::empty() {
  get_next_level();
  return queues[current_level].empty();
}

template <class T, class labelT>
ERROR hqws(T *in, labelT *out, size_t w, size_t h,
           const structuring_element &se, bool ignore_label_limit) {
  return hqws(in, out, w, h, 1, se, ignore_label_limit);
}

template <class T, class labelT>
ERROR hqws(T *in, labelT *out, size_t w_, size_t h_, size_t d_,
           const structuring_element &se, bool ignore_label_limit) {
  domain d(w_, h_, d_);
  T *tmp = new T[d.vol]();

  minima(in, tmp, d, se);
  size_t nbr_bh = label(tmp, out, d, se);
  delete[] tmp;
  if (!ignore_label_limit) {
    if (nbr_bh > std::numeric_limits<labelT>::max()) {
      return ERROR_LABEL;
    }
  }

  ERROR result = _hqws(in, out, d, se);

  return result;
}
template <class T, class labelT>
ERROR hqws(T *in, labelT *markers, labelT *out, size_t w, size_t h,
           const structuring_element &se) {
  return hqws(in, markers, out, w, h, 1, se);
}

template <class T, class labelT>
ERROR hqws(T *in, labelT *markers, labelT *out, size_t w, size_t h, size_t d,
           const structuring_element &se) {
#ifdef _OPENMP
  int nbr_threads = omp_get_max_threads();
#endif // _OPENMP

#ifdef _OPENMP
#pragma omp parallel for simd num_threads(nbr_threads)
#endif // _OPENMP
  for (size_t i = 0; i < w * h * d; ++i) {
    out[i] = markers[i];
  }

  ERROR result = _hqws(in, out, domain(w, h, d), se);
  return result;
}

#ifndef WS
#define WS std::numeric_limits<labelT>::max()
#endif

template <class T, class labelT>
ERROR _hqws(T *in, labelT *out, const domain &d, const structuring_element &se) {
#define HQ_LABELED 1
#define HQ_QUEUED 2
#define HQ_CANDIDATE std::numeric_limits<T>::max() - 1
#define HQ_WS std::numeric_limits<T>::max()

  T *status = new T[d.vol]();

  hierarchical_queue<T> hq;

// Benchmarks are realised with openmp time function...
#ifdef _OPENMP
#ifdef BENCH
  DECLARE_TIMING(hqws);
  START_TIMING(hqws);
#endif // BENCH
#endif // _OPENMP

  for (size_t p = 0; p < d.vol; ++p) {
    if (out[p] > 0 && out[p] < WS) {
      hq.push(in[p], p);
      status[p] = HQ_LABELED;
    } else
      status[p] = HQ_CANDIDATE;
  }

  size_t p;
  while (!hq.empty()) {
    p = hq.pop();

    status[p] = HQ_LABELED;

    // Processing label.
    ITERATE_NEIGHBORS_BEGIN(p, d, se) {
      if (status[_p] == HQ_LABELED) {
        if (out[p] == 0)
          out[p] = out[_p];
        else if (out[p] != out[_p])
          status[p] = HQ_WS;
      }
    }
    ITERATE_NEIGHBORS_END

    // Processing neighbors.
    if (status[p] != HQ_WS) {
      ITERATE_NEIGHBORS_BEGIN(p, d, se) {
        if (status[_p] == HQ_CANDIDATE) {
          hq.push(in[_p], _p);
          status[_p] = HQ_QUEUED;
        }
      }
      ITERATE_NEIGHBORS_END
    }
  }

#ifdef _OPENMP
  int nbr_threads = omp_get_max_threads();
#endif // _OPENMP

#ifdef _OPENMP
#pragma omp parallel for simd num_threads(nbr_threads)
#endif // _OPENMP
  for (size_t i = 0; i < d.vol; ++i) {
    if (status[i] == HQ_WS || status[i] == HQ_CANDIDATE) {
      out[i] = WS;
    }
  }

#ifdef _OPENMP
#ifdef BENCH
  STOP_TIMING(hqws);
  std::cout << GET_TIMING(hqws) << std::endl;
#endif // BENCH
#endif // _OPENMP

  delete[] status;

  return RES_OK;
}

#endif // _HIERARCHICAL_QUEUE_H_
