/*
 * Copyright 2018, Theodore Chabardes
 * <theodore.chabardes@mines-paristech.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef _IO_PNG_RAW_HPP_
#define _IO_PNG_RAW_HPP_

#include <iostream>
#include <limits.h>
#include <math.h>
#include <png.h>
#include <stddef.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// PNG signature length.
#define PNG_SIG_LEN 8

// local error handler.
static int io_png_read_abort(FILE *fp, png_structp *png_ptr,
                             png_infop *info_ptr) {
  png_destroy_read_struct(png_ptr, info_ptr, NULL);
  if (fp != NULL)
    fclose(fp);
  return -1;
}

// Internal function to check endianness.
bool is_big_endian(void) {
  union {
    uint32_t i;
    char c[4];
  } bint = {0x01020304};
  return bint.c[0] == 1;
}

// Read png file into an array.
template <class T>
static int io_png_read(T *&im, const char *path, size_t *_s_x, size_t *_s_y) {
  FILE *fp = fopen(path, "rb");
  png_byte magic[PNG_SIG_LEN];
  png_structp png_ptr;
  png_infop info_ptr;
  int expected_bit_depth = sizeof(T) * 8;

  // Header
  int bit_depth;
  int nbr_bytes;
  int png_transform = PNG_TRANSFORM_STRIP_ALPHA;

  // Parameters check.
  if (fp == NULL || path == NULL || _s_x == NULL || _s_y == NULL) {
    std::cerr << "io_png_raw: uncorrect parameters." << std::endl;
    return -1;
  }
  // Opening the png file.
  if (fread(magic, 1, PNG_SIG_LEN, fp) != PNG_SIG_LEN) {
    std::cerr
        << "io_png_raw: error while reading the signature of the png file."
        << std::endl;
    return io_png_read_abort(fp, NULL, NULL);
  }
  // Creating the reading structures.
  png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  if (png_ptr == NULL) {
    std::cerr << "io_png_raw: error while creating reading structure."
              << std::endl;
    return io_png_read_abort(fp, &png_ptr, NULL);
  }
  // Reading the informations.
  info_ptr = png_create_info_struct(png_ptr);
  if (info_ptr == NULL) {
    std::cerr << "io_png_raw: error while creating info structure."
              << std::endl;
    return io_png_read_abort(fp, &png_ptr, NULL);
  }
  // Checking the magic bits.
  if (!png_check_sig(magic, sizeof(magic))) {
    std::cerr << "io_png_raw: incorrect signature." << std::endl;
    return io_png_read_abort(fp, NULL, NULL);
  }
  if (setjmp(png_jmpbuf(png_ptr))) {
    std::cerr << "io_png_raw: unknown errors." << std::endl;
    return io_png_read_abort(fp, &png_ptr, &info_ptr);
  }
  png_init_io(png_ptr, fp);
  png_set_sig_bytes(png_ptr, PNG_SIG_LEN);

  // No alpha-channel.
  png_transform |= PNG_TRANSFORM_STRIP_ALPHA;
  // 1,2 and 4-bit samples to bytes.
  png_transform |= PNG_TRANSFORM_PACKING;
  // Endianness
  if (expected_bit_depth == 16 && !is_big_endian())
    png_transform |= PNG_TRANSFORM_SWAP_ENDIAN;

  // Reading the raw data.
  png_read_png(png_ptr, info_ptr, png_transform, NULL);

  bit_depth = png_get_bit_depth(png_ptr, info_ptr);
  if (expected_bit_depth != bit_depth) {
    std::cerr << "io_png_raw: image is not of expected depth." << std::endl;
    return io_png_read_abort(fp, &png_ptr, &info_ptr);
  }

  png_bytepp row_pointers = png_get_rows(png_ptr, info_ptr);
  size_t s_x = (size_t)png_get_image_width(png_ptr, info_ptr);
  size_t s_y = (size_t)png_get_image_height(png_ptr, info_ptr);
  size_t s_c = (size_t)png_get_channels(png_ptr, info_ptr);
  if (s_c != 1) {
    std::cerr << "io_png_raw: something went wrong..." << std::endl;
    return io_png_read_abort(fp, &png_ptr, &info_ptr);
  }
  *_s_x = s_x;
  *_s_y = s_y;
  nbr_bytes = (bit_depth <= 8) ? 1 : 2;

  im = new T[s_x * s_y];
  if (im == NULL) {
    std::cerr << "io_png_raw: error while allocating the image array."
              << std::endl;
    return io_png_read_abort(fp, &png_ptr, &info_ptr);
  }
  for (size_t j = 0; j < s_y; ++j) {
    memcpy(im + j * s_x, (void *)(row_pointers[j]), s_x * nbr_bytes);
  }

  png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
  fclose(fp);

  return 0;
}

static int io_png_write_abort(FILE *fp, png_structp *png_ptr,
                              png_infop *info_ptr) {
  png_destroy_write_struct(png_ptr, info_ptr);
  if (NULL != fp)
    (void)fclose(fp);
  return -1;
}

template <class T>
int io_png_write(const char *path, const T *data, size_t s_x, size_t s_y) {

  FILE *fp = fopen(path, "wb");
  png_structp png_ptr;
  png_infop info_ptr;
  int nbr_bits = sizeof(T) * 8;

  // Header
  int bit_depth;
  int color_type;
  int interlace_type;
  int compression;
  int filter;
  int nbr_bytes = nbr_bits / 8;

  if (fp == NULL || path == NULL || data == NULL || s_x == 0 || s_y == 0)
    return -1;

  png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  if (png_ptr == NULL) {
    std::cerr
        << "io_png_raw: error while reading the signature of the png file."
        << std::endl;
    return io_png_write_abort(fp, &png_ptr, NULL);
  }
  info_ptr = png_create_info_struct(png_ptr);
  if (info_ptr == NULL) {
    std::cerr << "io_png_raw: error while creating reading structure."
              << std::endl;
    return io_png_write_abort(fp, &png_ptr, NULL);
  }

  if (setjmp(png_jmpbuf(png_ptr))) {
    std::cerr << "io_png_raw: Unknown errors." << std::endl;
    return io_png_write_abort(fp, &png_ptr, &info_ptr);
  }

  png_init_io(png_ptr, fp);

  bit_depth = nbr_bytes * 8;
  color_type = PNG_COLOR_TYPE_GRAY;
  interlace_type = PNG_INTERLACE_ADAM7;
  compression = PNG_COMPRESSION_TYPE_BASE;
  filter = PNG_FILTER_TYPE_DEFAULT;

  png_set_IHDR(png_ptr, info_ptr, (png_uint_32)s_x, (png_uint_32)s_y, bit_depth,
               color_type, interlace_type, compression, filter);

  png_bytepp row_pointers;
  if ((row_pointers = (png_bytepp)malloc(s_y * sizeof(png_bytep))) == NULL) {
    std::cerr << "io_png_raw: error while allocating the row pointers."
              << std::endl;
    return io_png_write_abort(fp, &png_ptr, &info_ptr);
  }
  for (size_t j = 0; j < s_y; ++j) {
    row_pointers[j] = (png_bytep)((uint8_t *)data + j * s_x * nbr_bytes);
  }
  // png_set_rows(png_ptr, info_ptr, row_pointers);

  // png_transform = PNG_TRANSFORM_IDENTITY;
  // if (nbr_bytes == 2 && !is_big_endian())
  //   png_transform = PNG_TRANSFORM_SWAP_ENDIAN;
  // png_write_png(png_ptr, info_ptr, png_transform, NULL);
    png_write_info(png_ptr, info_ptr);
    png_write_image(png_ptr, row_pointers);
    png_write_end(png_ptr, info_ptr);

    free(row_pointers);
  png_destroy_write_struct(&png_ptr, &info_ptr);
  if (NULL != fp)
    (void)fclose(fp);
  return 0;
}

template <class T>
static int io_raw_read(T *&im, const char *path, const size_t &s_x,
                       const size_t &s_y, const size_t &s_z) {
  int nbr_bits = sizeof(T) * 8;
  int nbr_bytes = nbr_bits / 8;

  FILE *fp = NULL;
  fp = fopen(path, "rb");
  if (fp == NULL) {
    std::cerr << "io_png_raw: error while opening file. (" << path << ")"
              << std::endl;
    return -1;
  }

  im = new T[s_x * s_y * s_z];

  if (fread((void *)im, nbr_bytes, s_x * s_y * s_z, fp) != s_x * s_y * s_z) {
    std::cerr << "io_png_raw: error while reading file. (" << path << ")"
              << std::endl;
    delete[] im;
    return -1;
  }
  fclose(fp);

  return 0;
}

template <class T>
int io_raw_write(const char *path, const T *data, size_t s_x, size_t s_y,
                 size_t s_z) {
  int nbr_bits = sizeof(T) * 8;
  int nbr_bytes = nbr_bits / 8;

  FILE *fp = NULL;
  fp = fopen(path, "wb");
  if (fp == NULL) {
    std::cerr << "io_png_raw: error while opening file. (" << path << ")"
              << std::endl;
    return -1;
  }

  fwrite((void *)data, nbr_bytes, s_x * s_y * s_z, fp);

  fclose(fp);

  return 0;
}

#endif /* !_IO_PNG_RAW_HPP_ */
