/*
 * Copyright 2018, Theodore Chabardes
 * <theodore.chabardes@mines-paristech.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef _LABEL_H_
#define _LABEL_H_

#include "connexity.hpp"
#include "neighbors.hpp"
#include <queue>

// One pass labelling for the smallest images.
// 0 means "not a minimum".
// A positive number i means this pixel is within the connected component
// labeled i.
template <class T1, class T2>
size_t slow_label(const T1 *in, T2 *out, const domain &d,
                  const structuring_element &se) {

#ifdef _OPENMP
  int nbr_threads = omp_get_max_threads();
#pragma omp parallel for simd num_threads(nbr_threads)
#endif // _OPENMP
       // Clearing out the output image.
  for (size_t i = 0; i < d.vol; ++i)
    out[i] = 0;

  T2 cur_label = 0;
  size_t nbr_labels = 0;
  std::queue<size_t> q;

  for (size_t i = 0; i < d.vol; ++i) {
    if (in[i] != 0 && out[i] == 0) {

      ++cur_label;
      ++nbr_labels;
      if (cur_label == std::numeric_limits<T2>::max() - 1)
        cur_label = 1;

      q.push(i);
      out[i] = cur_label;

      while (!q.empty()) {
        size_t p = q.front();
        q.pop();

        ITERATE_NEIGHBORS_BEGIN(p, d, se) {
          if (in[_p] != 0 && out[_p] == 0) {
            out[_p] = cur_label;
            q.push(_p);
          }
        }
        ITERATE_NEIGHBORS_END
      }
    }
  }

  return nbr_labels;
}

template <class T1, class T2>
size_t label(const T1 *in, T2 *out, const domain &d,
             const structuring_element &se) {
  // No optimisation on labeling. See https://github.com/prittt/YACCLAB for
  // potential candidates for optimised CCL algorithms.
  return slow_label(in, out, d, se);
}

#endif // LABEL_H
