/*
 * Copyright 2018, Theodore Chabardes
 * <theodore.chabardes@mines-paristech.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef _MINIMA_H_
#define _MINIMA_H_

#include "connexity.hpp"
#include "neighbors.hpp"
#include <cstdlib>
#ifdef _OPENMP
#include <omp.h>
#endif // _OPENMP
#include <queue>

// Optimised minima computation, see section (3.4).
// Returns an image with values within {0,1} for all pixels:
//    - 1 for minima.
//    - 0 for others
template <class T1, class T2>
void minima(const T1 *in, T2 *out, const domain &d,
            const structuring_element &se) {

#ifdef _OPENMP
  int nbr_threads = omp_get_max_threads();
#endif // _OPENMP

#ifdef _OPENMP
#pragma omp parallel num_threads(nbr_threads)
#endif // _OPENMP
  {
    bool is_level_neighbor;
    std::queue<size_t> q;

#ifdef _OPENMP
#pragma omp for simd
#endif // _OPENMP
    for (size_t p = 0; p < d.vol; ++p)
      out[p] = 1;

#ifdef _OPENMP
#pragma omp for
#endif // _OPENMP

    // We start by detecting all descending border of plateaus.
    for (size_t p = 0; p < d.vol; ++p) {
      is_level_neighbor = false;

      ITERATE_NEIGHBORS_BEGIN(p, d, se) {
        if (in[_p] < in[p])
          out[p] = 0;
        if (in[_p] == in[p])
          is_level_neighbor = true;
      }
      ITERATE_NEIGHBORS_END

      if (is_level_neighbor && out[p] == 0)
        q.push(p);
    }

    // And we propagate the 0 value to the rest of the plateau.
    while (!q.empty()) {
      size_t p = q.front();
      q.pop();

      ITERATE_NEIGHBORS_BEGIN(p, d, se) {
        if (in[_p] == in[p] && out[_p] != 0) {
          out[_p] = 0;
          q.push(_p);
        }
      }
      ITERATE_NEIGHBORS_END
    }
  }
}

#endif // _MINIMA_H_
