/*
 * Copyright (c) 2010-2011, Nicolas Limare <nicolas.limare@cmla.ens-cachan.fr>
 * All rights reserved.
 * Copyright (c) 2019, Pascal Monasse <pascal.monasse@enpc.fr>
 * (io_png_write_color)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under, at your option, the terms of the GNU General Public
 * License as published by the Free Software Foundation, either
 * version 3 of the License, or (at your option) any later version, or
 * the terms of the simplified BSD license.
 *
 * You should have received a copy of these licenses along this
 * program. If not, see <http://www.gnu.org/licenses/> and
 * <http://www.opensource.org/licenses/bsd-license.html>.
 */


#ifndef OUTPUT_COLOR_HPP
#define OUTPUT_COLOR_HPP

#include <map>
#include <limits>
#include <png.h>
#include <cstring>

/**
 * local error structure
 * see http://www.libpng.org/pub/png/book/chapter14.htmlpointer
 */
typedef struct _io_png_err_s {
    jmp_buf jmpbuf;
} _io_png_err_t;

/**
 * local error handler
 * see http://www.libpng.org/pub/png/book/chapter14.htmlpointer
 */
static void _io_png_err_hdl(png_structp png_ptr, png_const_charp msg)
{
    _io_png_err_t *err_ptr;

    fprintf(stderr, "libpng error: %s\n", msg);

    err_ptr = (_io_png_err_t *) png_get_error_ptr(png_ptr);
    if (NULL == png_ptr) {
        fprintf(stderr, "fatal unrecoverable error, terminating\n");
        fflush(stderr);
        abort();
    }

    longjmp(err_ptr->jmpbuf, 1);
}

/**
 * @brief internal function used to cleanup the memory when
 * png_write_raw() fails
 *
 * @param fp file pointer to close, ignored if NULL
 * @param idata, row_pointers arrays to free, ignored if NULL
 * @param png_ptr_p, info_ptr_p, pointers to PNG structure pointers,
 *        ignored if NULL
 * @return -1
 */
static int _io_png_write_abort(FILE * fp,
                               png_bytep * row_pointers,
                               png_structp * png_ptr_p,
                               png_infop * info_ptr_p)
{
    png_destroy_write_struct(png_ptr_p, info_ptr_p);
    if (NULL != row_pointers)
        free(row_pointers);
    if (NULL != fp && stdout != fp)
        (void) fclose(fp);
    return -1;
}

/**
 * @brief internal function used to write a byte array as a PNG file
 *
 * The PNG file is written as a 8bit image file, interlaced,
 * truecolor. Depending on the number of channels, the color model is
 * gray, gray+alpha, rgb, rgb+alpha.
 *
 * @todo handle 16bit
 *
 * @param fname PNG file name, "-" means stdout
 * @param data interlaced RGB image byte array
 * @param nx, ny number of columns, lines
 * @return 0 if everything OK, -1 if an error occured
 */
static int _io_png_write_color(const char *fname, unsigned char *data,
                               size_t nx, size_t ny)
{
    png_structp png_ptr;
    png_infop info_ptr;
    png_bytep *row_pointers = NULL;
    png_byte bit_depth;
    /* volatile: because of setjmp/longjmp */
    FILE *volatile fp;
    int color_type, interlace, compression, filter;
    size_t j;
    /* error structure */
    _io_png_err_t err;

    /* parameters check */
    if (0 >= nx || 0 >= ny)
        return -1;
    if (NULL == fname || NULL == data)
        return -1;

    /* open the PNG output file */
    if (0 == strcmp(fname, "-"))
        fp = stdout;
    else if (NULL == (fp = fopen(fname, "wb")))
        return -1;

    if (NULL == (row_pointers = (png_bytep *) malloc(ny * sizeof(png_bytep))))
        return _io_png_write_abort(fp, NULL, NULL, NULL);

    /*
     * create and initialize the png_struct
     * with local error handling
     */
    if (NULL == (png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING,
                                                   &err, &_io_png_err_hdl,
                                                   NULL)))
        return _io_png_write_abort(fp, row_pointers, NULL, NULL);

    /* allocate/initialize the memory for image information */
    if (NULL == (info_ptr = png_create_info_struct(png_ptr)))
        return _io_png_write_abort(fp, row_pointers, &png_ptr, NULL);

    /* handle write errors */
    if (0 != setjmp(err.jmpbuf))
        /* if we get here, we had a problem writing to the file */
        return _io_png_write_abort(fp, row_pointers, &png_ptr,
                                   &info_ptr);

    /* set up the input control using standard C streams */
    png_init_io(png_ptr, fp);

    /* set image informations */
    bit_depth = 8;
    color_type = PNG_COLOR_TYPE_RGB;

    interlace = PNG_INTERLACE_ADAM7;
    compression = PNG_COMPRESSION_TYPE_BASE;
    filter = PNG_FILTER_TYPE_BASE;

    /* set image header */
    png_set_IHDR(png_ptr, info_ptr, (png_uint_32) nx, (png_uint_32) ny,
                 bit_depth, color_type, interlace, compression, filter);
    /* TODO : significant bit (sBIT), gamma (gAMA), comments (text) chunks */
    png_write_info(png_ptr, info_ptr);

    /* set row pointers */
    for (j = 0; j < ny; j++)
        row_pointers[j] = data + (size_t) (3*nx * j);

    /* write out the entire image and end it */
    png_write_image(png_ptr, row_pointers);
    png_write_end(png_ptr, info_ptr);

    /* clean up and free any memory allocated, close the file */
    (void) _io_png_write_abort(fp, row_pointers, &png_ptr, &info_ptr);

    return 0;
}

// --- Output basins with random colors, watershed in white ---

struct color { unsigned char r,g,b; };

template <class T>
int io_png_write_color(const char *path, const T *data, size_t s_x,size_t s_y){
    unsigned char* out = new unsigned char[s_x*s_y*3];
    std::map<T,color> lut;
    color white = {255,255,255};
    lut[std::numeric_limits<T>::max()] = white;
    unsigned char* p=out;
    for(size_t i=0; i<s_x*s_y; i++) {
        color c;
        typename std::map<T,color>::iterator it = lut.find(data[i]);
        if(it == lut.end()) {
            c.r = rand(); c.g = rand(); c.b = rand();
            lut[data[i]] = c;
        } else
            c = it->second;
        p[0] = c.r; p[1] = c.g; p[2] = c.b;
        p += 3;
    }
    int ret = _io_png_write_color(path, out, s_x, s_y);
    delete [] out;
    return ret;
}

#endif
