/*
 * Copyright 2018, Theodore Chabardes
 * <theodore.chabardes@mines-paristech.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef _PARAWS_H_
#define _PARAWS_H_

enum ERROR { ERROR_LABEL = -2, ERROR_IN, RES_OK };

#include "debug.hpp"
#include <iostream>
#include <limits>
#ifdef _OPENMP
#include <omp.h>
#endif // _OPENMP
#include "concurrent_queue.hpp"
#include "connexity.hpp"
#include "counts.hpp"
#include "distance.hpp"
#include "flood.hpp"
#include "label.hpp"
#include "minima.hpp"
#include "plateau.hpp"
#ifdef BENCH
#include "bench.hpp"
#endif // BENCH
#include <queue>
#include <utility>

// 3D version with markers.
template <class T, class labelT>
ERROR paraws(T *in, labelT *markers, labelT *out, size_t w_, size_t h_,
             size_t d_, const structuring_element &se,
             bool ignore_label_limit) {
  domain d(w_, h_, d_);
  T *tmp = new T[d.vol]();
  ushort *bh_sets = new ushort[d.vol]();

#ifdef _OPENMP
  int nbr_threads = omp_get_max_threads();
#endif // _OPENMP

  // The relief is flooded from the markers, while the minima are marked as
  // buttonholes.
#ifdef _OPENMP
#pragma omp parallel for simd num_threads(nbr_threads)
#endif // _OPENMP
  for (size_t i = 0; i < d.vol; ++i) {
    // The relief is pierced at markers.
    if (markers[i] > 0)
      in[i] = 0;
    else if (in[i] == 0)
      in[i] = 1;
  }

  minima(in, tmp, d, se);
  // Markers are excluded, and therefore not labelled as buttonhole.
#ifdef _OPENMP
#pragma omp parallel for simd num_threads(nbr_threads)
#endif // _OPENMP
  for (size_t i = 0; i < d.vol; ++i) {
    if (markers[i] > 0) {
      tmp[i] = 0;
    }
  }
  // We label the buttonhole.
  size_t nbr_bh = label(tmp, bh_sets, d, se);
  delete[] tmp;
  if (!ignore_label_limit) {
    if (nbr_bh > std::numeric_limits<ushort>::max()) {
      std::cerr << "Label overflow for buttonholes: " << nbr_bh << std::endl;
      delete[] bh_sets;
      return ERROR_LABEL;
    }
  }

#ifdef _OPENMP
#pragma omp parallel num_threads(nbr_threads)
#endif // _OPENMP
  {
#ifdef _OPENMP
#pragma omp for
#endif // _OPENMP
    for (size_t i = 0; i < d.vol; ++i) {
      if (markers[i] > 0) {
        out[i] = markers[i];
      } else if (bh_sets[i] > 0) {
        // Minima become buttonholes.
        out[i] = BETA;
      } else
        // Everything else have no label.
        out[i] = 0;
    }
  }

  ERROR result = _paraws(in, bh_sets, out, nbr_bh, d, se);

  // Free memory
  delete[] bh_sets;

  return result;
}

// 2D version with markers.
template <class T, class labelT>
ERROR paraws(T *in, labelT *markers, labelT *out, size_t w, size_t h,
             const structuring_element &se, bool ignore_label_limit) {
  return paraws(in, markers, out, w, h, 1, se, ignore_label_limit);
}

// 3D version without marker.
template <class T, class labelT>
ERROR paraws(T *in, labelT *out, size_t w_, size_t h_, size_t d_,
             const structuring_element &se, bool ignore_label_limit) {
  domain d(w_, h_, d_);
  T *tmp = new T[d.vol];

  // The relief is flooded from the minima.
  minima(in, tmp, d, se);
  size_t nbr_lbl = label(tmp, out, d, se);
  delete[] tmp;
  if (!ignore_label_limit) {
    if (nbr_lbl > std::numeric_limits<labelT>::max() - 1) {
      std::cerr << "Too many basins: " << nbr_lbl << std::endl;
      return ERROR_LABEL;
    }
  }

  ushort *bh_sets = new ushort[d.vol]();

  ERROR result = _paraws(in, bh_sets, out, 0, d, se);

  // Free memory
  delete[] bh_sets;

  return result;
}

// 2D version without marker.
template <class T, class labelT>
ERROR paraws(T *in, labelT *out, size_t w, size_t h,
             const structuring_element &se, bool ignore_label_limit) {
  return paraws(in, out, w, h, 1, se, ignore_label_limit);
}

// Main function.
template <class T, class labelT>
ERROR _paraws(T *in, ushort *bh_sets, labelT *out, size_t nbr_bh,
              const domain &d, const structuring_element &se) {

  // This stores the geodesic distance.
  ushort *homotopy = new ushort[d.vol]();
  // This stores the number of processed predecessors.
  uint8_t *counts = new uint8_t[d.vol]();

  BHData<ushort> buttonholes(nbr_bh);

#ifdef _OPENMP
  int nbr_threads = omp_get_max_threads();
#endif // _OPENMP

  moodycamel::ConcurrentQueue<size_t> tasksPool;

#ifdef _OPENMP
#ifdef BENCH
  DECLARE_TIMING(paraws);
  START_TIMING(paraws);
#endif // BENCH
#endif // _OPENMP

#ifdef _OPENMP
#pragma omp parallel num_threads(nbr_threads) shared(buttonholes, tasksPool)
#endif // _OPENMP
  {
    std::queue<size_t> q;

    // Queueing descending borders of plateau.
    detect_plateau(in, homotopy, q, d, se);

    // Queueing markers
#ifdef _OPENMP
#pragma omp for
#endif // _OPENMP
    for (size_t i = 0; i < d.vol; ++i) {
      if (out[i] > 0 && out[i] <= BETA) {
        q.push(i);
        homotopy[i] = 0;
      }
    }

    // Computing the modified relief on plateau.
    geodesic_distance(in, q, homotopy, d, se);

#ifdef _OPENMP
#pragma omp barrier
#pragma omp for
#endif // _OPENMP
    for (size_t i = 0; i < d.vol; ++i) {
      // Those are not in a plateau
      if (homotopy[i] == std::numeric_limits<unsigned short>::max())
        homotopy[i] = 0;
      ++homotopy[i];
      if (out[i] > 0 && out[i] <= BETA)
        homotopy[i] = 0;
    }

    create_counts_prec(in, homotopy, counts, d, se);

    size_t p;
    bool found;
    std::vector<size_t> to_be_queued;

#ifdef _OPENMP
#pragma omp for nowait
#endif // _OPENMP
    for (size_t i = 0; i < d.vol; ++i) {
      if (homotopy[i] == 0) {
        flood_point(i, buttonholes, in, homotopy, counts, bh_sets, out, d, se,
                   to_be_queued, lower_neighbor);
        counts[i] = 0;
      }
    }

    // Queueing multiple points at once is more efficient.
    tasksPool.enqueue_bulk(to_be_queued.data(), to_be_queued.size());
    to_be_queued.clear();
    do {
      found = tasksPool.try_dequeue(p);
      if (found) {
        flood_point(p, buttonholes, in, homotopy, counts, bh_sets, out, d, se,
                   to_be_queued, lower_neighbor_abs);

      } else if (to_be_queued.size() > 0) {
        tasksPool.enqueue_bulk(to_be_queued.data(), to_be_queued.size());
        to_be_queued.clear();
        found = true;
      }
    } while (found);

#ifdef _OPENMP
#pragma omp for
#endif // _OPENMP
    for (size_t i = 0; i < d.vol; ++i) {
      if (bh_sets[i] > 0) {
        bh_sets[i] = buttonholes.set.Find(bh_sets[i]);
      }
    }

#ifdef _OPENMP
#pragma omp for
#endif // _OPENMP
    for (size_t i = 1; i < buttonholes.size(); ++i) {
      ushort root = buttonholes.set.Find(i);
      if (root != i)
        buttonholes.BCL[root]->fuse_sync(buttonholes.BCL[i]);
    }

#ifdef _OPENMP
#pragma omp for
#endif // _OPENMP
    for (size_t i = 1; i < buttonholes.size(); ++i) {
      ushort root = buttonholes.set.Find(i);
      if (root == i)
        flood_buttonhole(root, buttonholes, in, homotopy, bh_sets, out, d, se);
    }

    // In some case, some BETA labels remain, such as buttonholes fully enclosed
    // by WS. Thus, those points receive WS as a label.
#ifdef _OPENMP
#pragma omp for simd nowait
#endif // _OPENMP
    for (size_t i = 0; i < d.vol; ++i) {
      if (out[i] == BETA)
        out[i] = WS;
    }
  }
#ifdef _OPENMP
#ifdef BENCH
  STOP_TIMING(paraws);
  std::cout << GET_TIMING(paraws) << std::endl;
#endif // BENCH
#endif // _OPENMP

  delete[] homotopy;
  delete[] counts;

  return RES_OK;
}

#endif // _PARAWS_H_
