/*
 * Copyright 2018, Theodore Chabardes
 * <theodore.chabardes@mines-paristech.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef _DETECT_PLATEAU_
#define _DETECT_PLATEAU_

#include "connexity.hpp"
#include <queue>

// Returns a binary image such as each pixel value is:
//   - 0 if it is in the descending border of a plateau.
//   - maximum of type unsigned short otherwise.
// It also stores the descending border of all plateaus in the queue q
template <class T>
void detect_plateau(const T *in, unsigned short *out, std::queue<size_t> &q, const domain d,
                    const structuring_element &se) {

  bool is_lvl_neighbor;
  bool is_descending;

#ifdef _OPENMP
#pragma omp for simd
#endif // _OPENMP
  for (size_t p = 0; p < d.vol; ++p) {
    out[p] = std::numeric_limits<unsigned short>::max();
  }

#ifdef _OPENMP
#pragma omp for
#endif // _OPENMP
  for (size_t p = 0; p < d.vol; ++p) {
    is_lvl_neighbor = false;
    is_descending = false;

    ITERATE_NEIGHBORS_BEGIN(p, d, se) {

      if (in[_p] == in[p])
        is_lvl_neighbor = true;
      if (in[_p] < in[p])
        is_descending = true;
    }
    ITERATE_NEIGHBORS_END

    if (is_lvl_neighbor && is_descending) {
      // One queue per thread, so this is thread-safe.
      q.push(p);
      out[p] = 0;
    }
  }
}

#endif // _DETECT_PLATEAU_
