/*
 * Copyright 2018, Theodore Chabardes
 * <theodore.chabardes@mines-paristech.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef _RULES_H_
#define _RULES_H_

#include "buttonhole.hpp"
#include "connexity.hpp"
#include "neighbors.hpp"
#include <cstdint>
#include <queue>

#ifndef WS
#define WS std::numeric_limits<labelT>::max()
#endif
#define BETA std::numeric_limits<labelT>::max() - 1

/*
 * This function computes the set Q(p).
 * CQl is the cardinal of Q(p).
 * Ql is the unique label if it exists.
 * return True if there is a beta label in the predecessors, else False.
 */

template <class T1, class labelT>
bool has_beta_pred(const size_t &p, const T1 *in, const unsigned short *homotopy, const labelT *out,
            const domain &d, const structuring_element &se) {

  ITERATE_NEIGHBORS_BEGIN(p, d, se) {

    // Special treatement of beta label.
    if (lower_neighbor_abs(_p, p, in, homotopy) && out[_p] == BETA) {
      return true;
    }
  }
  ITERATE_NEIGHBORS_END
  return false;
}

template <class T1, class labelT>
void get_Ql(const size_t &p, const T1 *in, const unsigned short *homotopy, const labelT *out,
            const domain &d, const structuring_element &se, uint8_t &CQl,
            labelT &Ql, bool (*funcRel) (size_t, size_t, const T1*, const unsigned short*)) {
  CQl = 0;
  Ql = 0;

  labelT label;

  ITERATE_NEIGHBORS_BEGIN(p, d, se) {

    if (funcRel (_p, p, in, homotopy) && out[_p] != 0 && out[_p] < BETA) {
        label = out[_p];
        if (CQl == 0) {
          CQl = 1;
          Ql = label;
        } else if (label != Ql){
          ++CQl;
        }  
    }
  }
  ITERATE_NEIGHBORS_END
}

/*
 * This function performs Union-Find on buttonholes.
 */
template <class T, class labelT>
void set_buttonhole(size_t p, BHData<labelT> &buttonholes, const T *in,
                      const unsigned short *homotopy, labelT *bh_sets, const domain &d,
                      const structuring_element &se) {
  labelT A = 0, B;
  ITERATE_NEIGHBORS_BEGIN(p, d, se) {
    if (lower_neighbor_abs(_p, p, in, homotopy) && bh_sets[_p] > 0) {
      B = buttonholes.set.Find(bh_sets[_p]);
      if (A == 0)
        A = B;
      else if (!buttonholes.set.Same(A, B))
        buttonholes.set.Union(A, B);
    }
  }
  ITERATE_NEIGHBORS_END

  if (A == 0) {
    bh_sets[p] = buttonholes.push_back();
  } else
    bh_sets[p] = A;
}

#endif // _RULES_H_
