/*
 * Copyright 2017, Theodore Chabardes
 * <theodore.chabardes@mines-paristech.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "../src/concurrent_union_find.hpp"
#include <limits>
#include <omp.h>
#include <stdlib.h>
#include <time.h>
#include <vector>

// Enumerating the differents operations possible.
enum UF_ACTION { UF_NONE, UF_NEWSET, UF_UNION, UF_FIND };

void run_operations(UF_ACTION &op, std::pair<uint16_t, uint16_t> &args,
                    concurrent_union_find<uint16_t> &solver) {
  if (op == UF_UNION) {
    solver.Union(args.first, args.second);
  } else if (op == UF_FIND) {
    solver.Find(args.first);
  }
}

int test_union_find_ts(void) {

  int nbr_threads = omp_get_max_threads();

  size_t nbr_operations = 3000;
  size_t nbr_sets = 1000;

  size_t nbr_pushback = 0;

  // Creating a series of operations.
  std::vector<UF_ACTION> operations =
      std::vector<UF_ACTION>(nbr_operations, UF_NONE);
  for (size_t i = 0; i < nbr_operations; ++i) {
    operations[i] = UF_ACTION(rand() % 2 + 2);
    if (operations[i] == UF_NEWSET) {
      nbr_pushback++;
    }
  }

  // A series of argument for the operations:
  //  - If the operation is a find, only one argument
  //  - If the operation is a union, two arguments.
  //  - If the operation is a new set, no argument.
  std::vector<std::pair<uint16_t, uint16_t>> arguments =
      std::vector<std::pair<uint16_t, uint16_t>>(
          nbr_operations, std::pair<uint16_t, uint16_t>(0, 0));
#pragma omp for
  for (size_t i = 0; i < nbr_operations; ++i) {
    arguments[i].first = rand() % nbr_sets + 1;
    arguments[i].second = rand() % nbr_sets + 1;
  }

  concurrent_union_find<uint16_t> solver_monothread;
  concurrent_union_find<uint16_t> solver_multithread;

  // Monothread version
  for (size_t i = 0; i < nbr_operations; ++i) {
    run_operations(operations[i], arguments[i], solver_monothread);
  }

  // Multithread version
#pragma omp parallel num_threads(nbr_threads)
  {
#pragma omp for
    for (size_t i = 0; i < nbr_operations; ++i) {
      run_operations(operations[i], arguments[i], solver_multithread);
    }
  }

  std::cout << (int)solver_monothread.get_number_of_subset(nbr_sets) << " "
            << (int)solver_multithread.get_number_of_subset(nbr_sets)
            << std::endl;
  if (solver_monothread.get_number_of_subset(nbr_sets) !=
      solver_multithread.get_number_of_subset(nbr_sets))
    return -1;

  return 0;
}

int main(void) {
  srand(time(NULL));

  for (size_t i = 0; i < 100; ++i) {
    if (test_union_find_ts() == -1) {
      std::cout << "Union-Find structure is not thread-safe." << std::endl;
      return -1;
    }
  }
  std::cout << "Union-Find structure is thread-safe." << std::endl;
  return 0;
}
